@@ -2,6 +2,7 @@ import unittest
 
 from pony.orm import *
 from pony.orm.tests.testutils import *
+from pony.py23compat import PYPY2
 
 db = Database('sqlite', ':memory:')
 
@@ -88,7 +89,8 @@ class TestSelectFromSelect(unittest.TestCase):
         self.assertEqual(db.last_sql.count('SELECT'), 1)
 
     @db_session
-    @raises_exception(ExprEvalError, "`s.scholarship > 0` raises NameError: name 's' is not defined")
+    @raises_exception(ExprEvalError, "`s.scholarship > 0` raises NameError: name 's' is not defined" if not PYPY2
+        else "`s.scholarship > 0` raises NameError: global name 's' is not defined")
     def test_7(self):  # test access to original query var name from the new query
         q = select(s.first_name for s in Student if s.scholarship < 500)
         q2 = select(x for x in q if s.scholarship > 0)
@@ -377,7 +379,7 @@ class TestSelectFromSelect(unittest.TestCase):
 
     @db_session
     def test_45(self):
-        q = select(s for s in Student).order_by(Student.first_name).limit(3, offset=1)
+        q = select(s for s in Student).order_by(Student.first_name, Student.id).limit(3, offset=1)
         q2 = select(s for s in q if s.age > 18).limit(2, offset=1)
         q3 = select(s.last_name for s in q2).limit(2, offset=1)
         self.assertEqual(set(q3), {'Brown'})
